import torch


class TaggableObject(torch.nn.Module):
    """This is the base class for all taggable objects.

    This class is used to add tags to objects and to search for objects with a specific tag.
    It is a torch.nn.Module, so that any object class inheriting from this class can be used
    as a torch.nn.Module to benefit from its functionality, e.g. `state_dict`.

    Attributes:
        tags (list): List of tags.
        tag_shortcuts (dict): Cache for faster search.
        device (str): Device on which the object is located. The default is "cpu".
    """

    def __init__(self, tag, device="cpu"):
        """Initialize the object.

        Args:
            tag (str): Tag to add to the object. It can also be a comma-separated string of multiple tags.
            device (str): Device on which the object is located.
        """
        self.device = device
        super().__init__()

        self.requires_grad_(False)
        self.tags = []
        self.clear_cache()

        #: dict of matrices that are evaluated on the fly
        self._mat_eval_dict = {}

        if tag is not None:
            self.add_tag(tag)
        self.add_tag(self.__class__.__name__)

    @property
    def tag(self):
        return self.tags[0]

    def has_module(self, tag):
        """Check if object has a module with a specific tag.

        Args:
            tag (str): Tag to search for.

        Returns:
            TaggableObject: The object with the tag. If no object is found, None is returned.

        Note: The returned object can be any object inheriting from TaggableObject.
        """
        return self[tag, 0] is not None

    def find_objects(self, key):
        """Find objects with a specific tag.

        This method should be overridden for deeper search.

        Args:
            key (str): Tag to search for.

        Returns:
            list: List of objects with the tag.
        """
        result = []
        return result

    def clear_cache(self):
        """Clear the tag cache for faster search.

        This method is called automatically when a tag is added or removed.
        """
        self.tag_shortcuts = {}

    def set_tag_attrs(self, tag, attr, value):
        """Set an attribute of all objects with a specific tag.

        Args:
            tag (str): Tag to search for.
            attr (str): Attribute to set.
            value (any): Value to set to the attribute.
        """
        for obj in self[tag]:
            setattr(obj, attr, value)

    def call_tag_functions(self, tag, attr, **args):
        """Call a function of all objects with a specific tag.

        Args:
            tag (str): Tag to search for.
            attr (str): Function to call.
            args (dict): Arguments to pass to the function.
        """
        for obj in self[tag]:
            getattr(obj, attr)(args)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            k = key[0]
            index = key[1]
        else:
            k = key
            index = None

        # cache
        if key in self.tag_shortcuts:
            result = self.tag_shortcuts[key]

            if index is not None:
                if len(result) > 0:
                    return result[index]
                else:
                    return None
            else:
                return result

        # normal search
        result = []
        if k in self.tags or (k is type and isinstance(self, k)):
            result.append(self)

        result += self.find_objects(k)

        if k not in self.tag_shortcuts:
            self.tag_shortcuts[k] = result

        if index is not None:
            if len(result) > 0:
                result = result[index]
            else:
                result = None

        return result

    def add_tag(self, tag):
        """Add tags to the object.

        Adds a tag only if object doesn't have the tag.
        Tags should be separated by ",".

        Args:
            tag (str): Tag string to add.
        """
        for subtag in tag.split(","):
            if subtag.strip() not in self.tags and len(subtag.strip()) > 0:
                self.tags.append(subtag.strip())



from pymonntorch.NetworkCore.Behavior import Behavior
from pymonntorch.NetworkBehavior.EulerEquationModules.Helper import (
    eq_split,
    remove_units,
)


class Variable(Behavior):
    def __init__(self, *args, eq=None, **kwargs):
        super().__init__(*args, eq=eq, **kwargs)

    def initialize(self, neurons):
        super().initialize(neurons)

        n = neurons

        eq_parts = eq_split(self.parameter("eq", None))

        if eq_parts[1] == "=" and len(eq_parts) >= 3:
            self.var_name = eq_parts[0]
        else:
            print("invalid formula")

        self.add_tag("Variable " + self.var_name)

        eq_parts = remove_units(eq_parts, 2)

        self.var_init = "".join(eq_parts[2:])

        setattr(n, self.var_name, neurons.vector() + eval(self.var_init))

        setattr(n, self.var_name + "_new", neurons.vector() + eval(self.var_init))

    def forward(self, n):
        setattr(
            n, self.var_name, getattr(n, self.var_name + "_new")
        )  # apply the new value to variable


class SynapseVariable(Behavior):
    def __init__(self, *args, eq=None, **kwargs):
        super().__init__(*args, eq=eq, **kwargs)

    def initialize(self, synapse):
        s = synapse

        eq_parts = eq_split(self.parameter("eq", None))

        if eq_parts[1] == "=" and len(eq_parts) >= 3:
            self.var_name = eq_parts[0]
        else:
            print("invalid formula")

        self.add_tag("SynapseVariable " + self.var_name)

        eq_parts = remove_units(eq_parts, 2)

        self.var_init = "".join(eq_parts[2:])

        setattr(s, self.var_name, synapse.matrix() + eval(self.var_init))

        setattr(s, self.var_name + "_new", synapse.matrix() + eval(self.var_init))

    def forward(self, s):
        setattr(
            s, self.var_name, getattr(s, self.var_name + "_new")
        )  # apply the new value to variable



from pymonntorch.NetworkCore.Behavior import Behavior
from pymonntorch.NetworkBehavior.EulerEquationModules.Helper import (
    eq_split,
    remove_units,
)
from sympy import symbols
import torch


class Equation(Behavior):
    def __init__(self, *args, eq=None, step_size="1*ms", **kwargs):
        super().__init__(*args, eq=eq, step_size=step_size, **kwargs)

    def initialize(self, neurons):
        super().initialize(neurons)
        n = neurons
        self.add_tag("EquationModule")
        self.step_size = self.parameter("step_size", "1*ms", neurons)
        eq_parts = eq_split(self.parameter("eq", None))

        if (
            eq_parts[0][0] == "d"
            and eq_parts[1] == "/"
            and eq_parts[2] == "dt"
            and eq_parts[3] == "="
        ):
            self.variable = eq_parts[0][1:]
            symbols("n." + self.variable)
        else:
            print("invalid equation")

        for i in range(4, len(eq_parts)):
            if eq_parts[i] in neurons.__dict__:
                eq_parts[i] = "n." + eq_parts[i]

        eq_parts = remove_units(eq_parts, 4)

        self.evaluation = (
            "n."
            + self.variable
            + "+("
            + "".join(eq_parts[4:])
            + ")*{}".format(neurons.clock_step_size)
        )

        self.compiled_evaluation = compile(self.evaluation, "<string>", "eval")

        print(self.evaluation)

    def forward(self, n):
        new = eval(self.compiled_evaluation)
        setattr(n, self.variable + "_new", torch.tensor(new, device=n.device))



"""
This module includes functions and classes to facilitate recording of network object \
variables through out the simulation time.
"""
import copy
import torch
from pymonntorch.NetworkCore.Behavior import Behavior


def get_Recorder(variable):
    """Returns a `Recorder` instance for the given variable.

    Args:
        variable (str): Name of the variable to record.

    Returns:
        Recorder: The `Recorder` object.
    """
    return Recorder(variables=[variable])


class Recorder(Behavior):
    """This is the base class to record variables of a network object.
    
    Args:
        variables (list of str): List of variable names to record. A variable should be of tensor type.
        gap_width (int): The intervals of time to record variables. The default is 0.
        tag (str): A tag name for the `Recorder` object. The default is None.
        max_length (int): The history buffer size. If `None`, the variables are recorded \
            for the whole simulation time. The default is None.
        auto_annotate (bool): This parameter specifies whether the variable names include the \
            network object prefix (neurons/synapse/n/s) or not. The default is True.
    """

    initialize_last = True
    visualization_module_outputs = []

    def __init__(
        self,
        *args,
        variables=None,
        gap_width=0,
        max_length=None,
        auto_annotate=True,
        tag=None,
        **kwargs,
    ):
        super().__init__(
            *args,
            variables=variables,
            gap_width=gap_width,
            max_length=max_length,
            auto_annotate=auto_annotate,
            tag=tag,
            **kwargs,
        )
        self.variables = {}

    def initialize(self, object):
        super().initialize(object)

        variables = self.parameter("variables", None)
        if variables is None:
            variables = self.parameter("arg_0", None)
            if variables is None:
                variables = []
        if isinstance(variables, str):
            variables = [variables]

        self.gap_width = self.parameter("gap_width", 0)
        self.max_length = self.parameter("max_length", None)
        self.auto_annotate = self.parameter("auto_annotate", True)
        self.counter = 0
        self.new_data_available = False
        self.compiled = {}

        self.add_variables(variables)
        self.reset()

    def add_variable(self, v):
        self.variables[v] = torch.tensor([], dtype=torch.bool, device=self.device)
        self.compiled[v] = None

    def add_variables(self, vars):
        for v in vars:
            self.add_variable(v)

    def find_objects(self, key):
        result = []
        if key in self.variables:
            result.append(self.variables[key])
        return result

    def reset(self):
        for v in self.variables:
            self.add_variable(v)

    def is_new_data_available(self):
        if self.new_data_available:
            self.new_data_available = False
            return True
        else:
            return False

    def get_data_v(self, variable, parent_obj):
        n = parent_obj  # used for eval string "n.x"
        s = parent_obj
        neurons = parent_obj
        synapse = parent_obj

        return copy.copy(eval(self.compiled[variable]))

    def save_data_v(self, data, variable):
        if self.variables[variable].dtype != data.dtype and torch.numel(
            self.variables[variable]
        ):
            print(
                f"WARNING: The recorder received new data with a different datatype({data.dtype}) from the previously recorded data({self.variables[variable].dtype})."
            )
        self.variables[variable] = torch.concat(
            (self.variables[variable], data.unsqueeze(0)), dim=0
        )

    def eq_split(self, eq, splitter):
        str = eq.replace(" ", "")
        parts = []
        str_buf = ""
        for s in str:
            if s in splitter:
                parts.append(str_buf)
                parts.append(s)
                str_buf = ""
            else:
                str_buf += s

        parts.append(str_buf)
        return parts

    def annotate_var_str(self, variable, parent_obj):
        splitter = [
            "*",
            "/",
            "+",
            "-",
            "%",
            ":",
            ";",
            "=",
            "!",
            "(",
            ")",
            "[",
            "]",
            "{",
            "}",
        ]
        annotated_var = ""
        for part in self.eq_split(variable, splitter):
            if hasattr(parent_obj, part):
                part = "n." + part
            annotated_var += part
        return annotated_var

    def forward(self, parent_obj):
        if parent_obj.recording:
            self.counter += 1

            if self.counter >= self.gap_width:
                self.new_data_available = True
                self.counter = 0

                for v in self.variables:
                    if self.compiled[v] is None:
                        if self.auto_annotate:
                            annotated_var = self.annotate_var_str(v, parent_obj)
                            self.compiled[v] = compile(
                                annotated_var, "<string>", "eval"
                            )
                        else:
                            self.compiled[v] = compile(v, "<string>", "eval")

                    data = self.get_data_v(v, parent_obj)
                    if data is not None:
                        self.save_data_v(data, v)

        if self.max_length is not None:
            self.cut_length(self.max_length)

    def cut_length(self, max_length):
        if max_length is not None:
            for v in self.variables:
                while len(self.variables[v]) > max_length:
                    self.variables[v] = self.variables[v][1:]

    def swapped(self, name):
        return self.swap(self.variables[name])

    def swap(self, x):
        return torch.swapaxes(x, 1, 0)

    def clear_recorder(self):
        print("clear")

        for v in self.variables:
            device = self.variables[v].device
            del self.variables[v]
            if device.type == "cuda":
                torch.cuda.empty_cache()
            self.variables[v] = torch.tensor([], dtype=torch.bool, device=device)


class EventRecorder(Recorder):
    """This class is used to record sparse boolean vectors over time more efficiently.
    It returns a tensor of (t, i) tuples where t indicates the time step and i is the 
    index of elements with value `True`. If the variable to record is not boolean itself,
    it is converted to one by assessing whether values are >0.
    
    Args:
        variables (list of str): List of variable names to record.
        gap_width (int): The intervals of time to record variables. The default is 0.
        tag (str): A tag name for the `Recorder` object. The default is None.
        max_length (int): The history buffer size. If `None`, the variables are recorded \
            for the whole simulation time. The default is None.
        auto_annotate (bool): This parameter specifies whether the variable names include the \
            network object prefix (neurons/synapse/n/s) or not. The default is True.
    """

    def find_objects(self, key):
        result = []
        if key in self.variables:
            result.append(self.variables[key])

        if type(key) is str and key[-2:] == ".t" and key[:-2] in self.variables:
            result.append(self.variables[key[:-2]][:, 0])

        if type(key) is str and key[-2:] == ".i" and key[:-2] in self.variables:
            result.append(self.variables[key[:-2]][:, 1])

        return result

    def get_data_v(self, variable, parent_obj):
        n = parent_obj  # used for eval string "n.x"
        s = parent_obj
        neurons = parent_obj
        synapse = parent_obj

        data = eval(self.compiled[variable])

        indices = torch.where(data != 0)
        iteration = torch.ones_like(indices[0]) * parent_obj.iteration

        return torch.stack((iteration, *indices), dim=1)

    def save_data_v(self, data, variable):
        self.variables[variable] = torch.concat([self.variables[variable], data])



import torch

from pymonntorch.NetworkCore.TaggableObject import *


class NetworkObject(TaggableObject):
    """This is the base class for all network objects.

    This class is used to treat network objects' behaviors and is a subclass of TaggableObject.

    Attributes:
        network (Network): The parent network object.
        behavior (list or dict): List or dictionary of behaviors.
        analysis_modules (list): List of analysis modules.
    """

    def __init__(self, network, tag=None, behavior=None, device=None):
        """Initialize the object.

        Args:
            tag (str): Tag to add to the object. It can also be a comma-separated string of multiple tags.
            network (Network): The parent network object.
            behavior (list or dict): List or dictionary of behaviors. If a dictionary is used, the keys must be integers.
            device (str): Device on which the object is located. If not provided object reside on network device.
        """
        if device is None:
            device = network.device
        if tag is None:
            tag = self.__class__.__name__ + "_" + str(len(network.Structures)+1)
        super().__init__(tag, device)

        self.network = network
        self.behavior = behavior if behavior is not None else {}
        if type(behavior) == list:
            self.behavior = dict(zip(range(len(behavior)), behavior))
        # self.behavior = torch.nn.ModuleDict(self.behavior)

        for b in self.behavior.values():
            if not hasattr(self, b.tags[0]):
                setattr(self, b.tags[0], b)

        for k, b in self.behavior.items():
            self.network._add_behavior_to_sorted_execution_list(k, self, b)

        for k in sorted(list(self.behavior.keys())):
            if self.behavior[k].initialize_on_init:
                self.behavior[k].initialize(self)

        self.analysis_modules = []

        self.sub_structures = []
        self.parent_structure = network
        if network != self:
            network.Structures.append(self)

        self.recording = True

    def add_sub_structure(self, new_structure):
        """add a new structure as sub structure

        Args:
            new_structure (NetworkObject): the new structure to add.
        """
        if new_structure.network == self.network:
            self.sub_structures.append(new_structure)
            new_structure.parent_structure = self
        else:
            print(f"ERROR: {new_structure.tags[0]} can not be a sub structure of {self.tags[0]} because of not mutual network.")

    def all_sub_structure(self):
        """Returns a list of all sub_structures.
        """
        result = []
        for sub_struc in self.sub_structures:
            result.extend(sub_struc.all_sub_structure())
        return result

    def add_behavior(self, key, behavior, initialize=True):
        """Add a single behavior to the network object.

        Args:
            key (str): Key to be used to access behavior.
            behavior (Behavior): Behavior to be added.
            initialize (bool): If true, behavior will be initialized. The default is True.

        Returns:
            Behavior: The behavior.
        """
        if key not in self.behavior:
            self.behavior[key] = behavior
            self.network._add_behavior_to_sorted_execution_list(
                key, self, self.behavior[key]
            )
            self.network.clear_tag_cache()
            if initialize:
                behavior.initialize(self)
                behavior.check_unused_attrs()
            return behavior
        else:
            raise Exception("Error: Key already exists." + str(key))

    def add_behaviors(self, behavior_dict):
        """Add multiple behaviors to the network object.

        Args:
            behavior_dict (dict): Dictionary of behaviors to be added. The keys must be integers.

        Returns:
            dict: The dictionary of behaviors.
        """
        for key, behavior in behavior_dict.items():
            self.add_behavior(key, behavior)
        return behavior_dict

    def remove_behavior(self, key_tag_behavior_or_type):
        """Remove behavior(s) from the network object.

        Args:
            key_tag_behavior_or_type (str, Behavior, or type): Key, tag, behavior object, or type of behavior to be removed.
        """
        remove_keys = []
        for key in self.behavior:
            b = self.behavior[key]
            if (
                key_tag_behavior_or_type == key
                or key_tag_behavior_or_type in b.tags
                or key_tag_behavior_or_type == b
                or key_tag_behavior_or_type == type(b)
            ):
                remove_keys.append(key)

        for key in remove_keys:
            b = self.behavior.pop(key)
            self.network._remove_behavior_from_sorted_execution_list(key, self, b)

    def set_behaviors(self, tag, enabled):
        """Set behaviors to be enabled or disabled.

        Args:
            tag (str): Tag of behaviors to be enabled or disabled.
            enabled (bool): If true, behaviors will be enabled. If false, behaviors will be disabled.
        """
        if enabled:
            print("activating", tag)
        else:
            print("deactivating", tag)
        for b in self[tag]:
            b.behavior_enabled = enabled

    def deactivate_behaviors(self, tag):
        """Disable behaviors.

        Args:
            tag (str): Tag of behaviors to be disabled.
        """
        self.set_behaviors(tag, False)

    def activate_behaviors(self, tag):
        """Enable behaviors.

        Args:
            tag (str): Tag of behaviors to be enabled.
        """
        self.set_behaviors(tag, True)

    def find_objects(self, key):
        """Find behaviors and analysis modules in the network object by key.

        Args:
            key (str): Key to be used to access behavior or analysis module.

        Returns:
            list: List of behaviors and analysis modules.
        """
        result = []

        if key in self.behavior:
            result.append(self.behavior[key])

        for bk in self.behavior:
            behavior = self.behavior[bk]
            result += behavior[key]

        for am in self.analysis_modules:
            result += am[key]

        return result

    def add_analysis_module(self, module):
        """Add an analysis module to the network object.

        Args:
            module (AnalysisModule): Analysis module to be added.
        """
        module._attach_and_initialize_(self)

    def get_all_analysis_module_results(self, tag, return_modules=False):
        """Get results from all analysis modules in the network object.

        Args:
            tag (str): Tag of analysis modules to be used.
            return_modules (bool): If true, the analysis modules will be returned. The default is False.

        Returns:
            dict: Dictionary of results.
        """
        result = {}
        modules = {}
        for module in self[tag]:
            module_results = module.get_results()
            for k in module_results:
                result[k] = module_results[k]
                modules[k] = module

        if return_modules:
            return result, modules
        else:
            return result

    def buffer_roll(self, mat, new=None, counter=False):
        """Shift the elements of a tensor to the right.

        Args:
            mat (torch.Tensor): Tensor to be shifted.
            new (int or float or bool or torch.Tensor): New element to be inserted at the beginning of the tensor. The default is None (i.e. the last of the buffer is repositoined at the first).
            counter (bool): If True, rolling is done in opposite direction, and the new element is added at the end.

        Returns:
            torch.Tensor: The shifted tensor.
        """
        mat = mat.roll(1 - (2 * counter), dims=0)

        if new is not None:
            mat[0 - counter] = new

        return mat

    def _get_mat(self, mode, dim, scale=None, density=None, plot=False, dtype=None):
        """Get a tensor with object's dimensionality.

        The tensor can be initialized in different modes. List of possible values for mode includes:
        - "random" or "rand" or "rnd" or "uniform": Uniformly distributed random numbers in range [0, 1).
        - "normal(mean=a, std=b)": Normally distributed random numbers with `a` as mean and `b` as standard derivation.
        - "ones": Tensor filled with ones.
        - "zeros": Tensor filled with zeros.
        - A single number: Tensor filled with that number.
        - You can also use any function from torch package for this purpose.

        Args:
            mode (str): Mode to be used to initialize tensor.
            dim (int or tuple of int): Dimensionality of the tensor.
            scale (float): Scale of the tensor. The default is None (i.e. No scaling is applied).
            density (float): Density of the tensor. The default is None (i.e. dense tensor).
            plot (bool): If true, the histogram of the tensor will be plotted. The default is False.
            dtype (str or type): Data type of the tensor. If None, `def_dtype` will be used.

        Returns:
            torch.Tensor: The initialized tensor."""

        dtype = self.def_dtype if dtype is None else dtype

        if mode not in self._mat_eval_dict:
            prefix = "torch."
            ev_str = mode

            if (
                ev_str == "random"
                or ev_str == "rand"
                or ev_str == "rnd"
                or ev_str == "uniform"
            ):
                ev_str = "rand"

            if type(ev_str) == int or type(ev_str) == float:
                ev_str = "ones()*" + str(ev_str)

            ev_str = prefix + ev_str
            if "(" not in ev_str and ")" not in ev_str:
                ev_str += "()"

            a1 = "size=dim,device=self.device,dtype=dtype)"

            # check for positional argument
            if ev_str[ev_str.index("(") + 1 : ev_str.index(")")].strip():
                a1 = "," + a1

            ev_str = ev_str.replace(")", a1)
            self._mat_eval_dict[mode] = compile(ev_str, "<string>", "eval")

        result = eval(self._mat_eval_dict[mode])

        if density is not None:
            if type(density) == int or type(density) == float:
                result = result * (torch.rand(dim, device=self.device) <= density)
            elif type(density) is torch.tensor:
                result = result * (
                    torch.rand(dim, device=self.device) <= density[:, None]
                )

        if scale is not None:
            result *= scale

        if plot:
            import matplotlib.pyplot as plt

            plt.hist(result.flatten().to("cpu"), bins=30)
            plt.show()

        return result

    def tensor(self, mode, dim, scale=None, density=None, dtype=None):
        """Get a tensor with desired dimensionality.

        The tensor can be initialized in different modes. List of possible values for mode includes:
        - "random" or "rand" or "rnd" or "uniform": Uniformly distributed random numbers in range [0, 1).
        - "normal(mean=a, std=b)": Normally distributed random numbers with `a` as mean and `b` as standard derivation.
        - "ones": Tensor filled with ones.
        - "zeros": Tensor filled with zeros.
        - A single number: Tensor filled with that number.
        - You can also use any function from torch package for this purpose.

        Args:
            mode (str): Mode to be used to initialize tensor.
            dim (int or tuple of int): Dimensionality of the tensor.
            scale (float): Scale of the tensor. The default is None (i.e. No scaling is applied).
            density (float): Density of the tensor. The default is None (i.e. dense tensor).
            dtype (str or type): Data type of the tensor. If None, `def_dtype` will be used.

        Returns:
            torch.Tensor: The initialized tensor."""
        return self._get_mat(
            mode=mode, dim=dim, scale=scale, density=density, dtype=dtype
        )

    def get_buffer_mat(self, dim, size, **kwargs):
        """Get a buffer of specific size with object's dimensionality.

        Args:
            dim (int or tuple of int): Dimensionality of the buffer.
            size (int): Size of the buffer.
            kwargs (dict): Keyword arguments to be passed to the initialization function.

        Returns:
            torch.Tensor: The buffer.
        """
        return (
            torch.cat([torch.zeros(dim, **kwargs) for _ in range(size)])
            .reshape(size, *dim)
            .to(self.device)
        )

    @property
    def iteration(self):
        """int: iteration number or time step."""
        return self.network._iteration

    @iteration.setter
    def iteration(self, iteration):
        if iteration >= 0 and type(iteration) is int:
            self.network._iteration = iteration
        else:
            print(
                "WARNING: Attempting to set an invalid value for iteration!\n Setting iteration to zero..."
            )
            self.network._iteration = 0

    @property
    def def_dtype(self):
        return self.network._def_dtype

    @def_dtype.setter
    def def_dtype(self, dtype):
        self.network._def_dtype = dtype



import sympy.physics.units as units
from sympy.physics.units import convert_to, second, seconds


def eq_split(eq):
    str = eq.replace(" ", "")
    parts = []
    str_buf = ""
    for s in str:
        if s in ["*", "/", "+", "-", "%", ":", ";", "=", "!", "(", ")", "[", "]"]:
            parts.append(str_buf)
            parts.append(s)
            str_buf = ""
        else:
            str_buf += s

    parts.append(str_buf)
    return parts


def remove_units(eq_parts, i):

    while i < len(eq_parts):

        if eq_parts[i] == "ms":

            t = float(
                convert_to(
                    eval(eq_parts[i - 2] + eq_parts[i - 1] + eq_parts[i]), seconds
                )
                / second
            )

            eq_parts[i - 2] = "{}".format(t)

            eq_parts.pop(i)
            eq_parts.pop(i - 1)
            i -= 2

        if eq_parts[i] in units.__dict__ or eq_parts[i] in myUnits:  # remove units
            eq_parts.pop(i)
            eq_parts.pop(i - 1)
            i -= 2
        i += 1

    return eq_parts


myUnits = ["mV"]



"""
This module includes functions and classes to handle structured NeuronGroups.
"""
import torch
from pymonntorch.NetworkCore.Behavior import Behavior


def vec_to_mat(vec, repeat_count):
    """
    Stack `repeat_count` numbers of the input vector on first dimension.

    Args:
        vec (torch.tensor): The vector to replicate.
        repeat_count (int): Number of replicas.

    Returns:
        torch.tensor: A matrix of the stacked `vec`s for `repeat_count` times.
    """
    return torch.stack([vec] * repeat_count, dim=0)


def vec_to_mat_transposed(vec, repeat_count):
    """
    Stack `repeat_count` numbers of the input vector on second dimension.

    Args:
        vec (torch.tensor): The vector to replicate.
        repeat_count (int): Number of replicas.

    Returns:
        torch.tensor: A matrix of the stacked `vec`s for `repeat_count` times.
    """
    return torch.stack([vec] * repeat_count, dim=1)


def rotation_matrix(axis, theta, dtype=torch.float):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given `axis` by `theta` radians.

    Args:
        axis (int): The axis.
        theta (float): The angle of rotation in radians.
        dtype (type): Type of the tensor values. The default is `torch.float`.

    Returns:
        torch.tensor: The rotation matrix.
    """
    axis = torch.tensor(axis, dtype=dtype)
    axis = axis / torch.sqrt(torch.dot(axis, axis))
    a = torch.cos(theta / 2.0)
    b, c, d = -axis * torch.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return torch.tensor(
        [
            [aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
            [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
            [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc],
        ]
    )


def get_squared_dim(n_neurons, depth=1):
    """
    Get the squared dimension of a square matrix that can hold `n_neurons`
    neurons, with `depth` layers. The neurons are structured in a square-like 
    manner.

    Args:
        n_neurons (int): The desired number of neurons in the `NeuronGroup`.
        depth (int): The depth of the structured `NeuronGroup`. The default is 1.

    Returns:
        NeuronDimension: An instance of `NeuronDimension` behavior with the given depth \
            and calculated height and depth.
    """
    divider = int(torch.trunc(torch.sqrt(torch.tensor(n_neurons))))

    while divider > 1 and n_neurons % divider != 0:
        divider -= 1

    width = divider
    height = torch.div(n_neurons, divider, rounding_mode="floor") if divider > 0 else 0

    print(f"Set population size to {depth}x{height}x{width}.")

    return NeuronDimension(width=width, height=height, depth=depth)


class NeuronDimension(Behavior):
    """
    The Behavior that defines structure for a `NeuronGroup`.
    It overrides the `size` variable of the `NeuronGroup` and adds `x`, `y`, and `z` vectors,
    as well as `width`, `height` and `depth` variables. The Behaviour is special, because its
    `initialize` function is executed when the `NeuronGroup` is created rather than
    when network.initialize() is called. The neurons are arranged in a 3-dimensional grid
    with size=width * height * depth. Because is overrides the `size` variable, it does not
    have to be added in the behaviour dictionary directly, but also indirectly
    ( `NeuronGroup(size=NeuronDimension(),...)` ). In this case it will be added to
    position 0 in the dictionary.

    Args:
        width (int): Width of the `NeuronGroup`. The default is 1.
        height (int): Height of the `NeuronGroup`. The default is 1.
        depth (int): Depth of the `NeuronGroup`. The default is 1.
    """

    initialize_on_init = True

    def __init__(
        self, *args, depth=1, height=1, width=1, input_patterns=None, **kwargs
    ):
        super().__init__(
            *args,
            depth=depth,
            height=height,
            width=width,
            input_patterns=input_patterns,
            **kwargs,
        )

    def set_position(self, width, height, depth):
        """Set the coordinate of neurons by setting vectors `x`, `y`, and `z`.

        Args:
            width (int): Width of the neurons.
            height (int): Height of the neurons.
            depth (int): Depth of the neurons.
        """
        self.neurons.x = (
            torch.arange(
                0,
                self.neurons.size,
                dtype=self.neurons.def_dtype,
                device=self.neurons.device,
            )
            % width
        )
        self.neurons.y = (
            torch.div(
                torch.arange(
                    0,
                    self.neurons.size,
                    dtype=self.neurons.def_dtype,
                    device=self.neurons.device,
                ),
                width,
                rounding_mode="floor",
            )
            % height
        )
        self.neurons.z = (
            torch.div(
                torch.arange(
                    0,
                    self.neurons.size,
                    dtype=self.neurons.def_dtype,
                    device=self.neurons.device,
                ),
                (width * height),
                rounding_mode="floor",
            )
            % depth
        )

    def get_area_mask(self, xmin=0, xmax=-1, ymin=0, ymax=-1, zmin=0, zmax=-1):
        """Returns a mask tensor with the same shape as the NeuronGroup
        with the given start and end points.

        Args:
            xmin (int): Start point in x axis. The default is 0.
            xmax (int): End point in x axis. The default is -1.
            ymin (int): Start point in y axis. The default is 0.
            ymax (int): End point in y axis. The default is -1.
            zmin (int): Start point in z axis. The default is 0.
            zmax (int): End point in z axis. The default is -1.

        Returns:
            torch.BoolTensor: The mask tensor.
        """
        if xmax < 1:
            xmax = self.width - xmax
        if ymax < 1:
            ymax = self.height - ymax
        if zmax < 1:
            zmax = self.depth - zmax

        result = torch.zeros(
            self.depth,
            self.height,
            self.width,
            dtype=torch.bool,
            device=self.neurons.device,
        )
        result[zmin:zmax, ymin:ymax, xmin:xmax] = True
        return result.flatten()

    def apply_pattern_transformation_function(self, transform_mat, hup, wup, depth):
        """Apply a transformation matrix on the neurons.

        Args:
            transform_mat (torch.tensor): The transformation matrix.
            hup (int): The height upperbound.
            wup (int): The width upperbound.
            depth (int): The depth.
        """
        big_x_mat = torch.stack(
            [torch.arange(wup)] * hup,
            dim=0,
            dtype=self.neurons.def_dtype,
            device=self.neurons.device,
        )
        big_x_mat = big_x_mat.repeat(depth, 1).flatten()

        big_y_mat = torch.repeat_interleave(
            torch.arange(wup, dtype=self.neurons.def_dtype, device=self.neurons.device),
            hup * depth,
        )

        self.neurons.x = big_x_mat[transform_mat]
        self.neurons.y = big_y_mat[transform_mat]

    def move(self, x=0, y=0, z=0):
        """Move the neurons in the 3D space.

        Args:
            x (int or float): The displacement in x axis.
            y (int or float): The displacement in y axis.
            z (int or float): The displacement in z axis.

        Returns:
            NeuronDimension: The modified `NeuronDimension`.
        """
        self.neurons.x += x
        self.neurons.y += y
        self.neurons.z += z
        return self

    def scale(self, x=1, y=1, z=1):
        """Scale the neuron grid in the 3D space.

        Args:
            x (int or float): The stretch in x axis.
            y (int or float): The stretch in y axis.
            z (int or float): The stretch in z axis.

        Returns:
            NeuronDimension: The modified `NeuronDimension`.
        """
        self.neurons.x *= x
        self.neurons.y *= y
        self.neurons.z *= z
        return self

    def noise(self, x_noise=1, y_noise=1, z_noise=1):
        """Apply random noise to neuron coordinates.

        Args:
            x (int or float): The noise bounds in x axis.
            y (int or float): The noise bounds in y axis.
            z (int or float): The noise bounds in z axis.

        Returns:
            NeuronDimension: The modified `NeuronDimension`.
        """
        self.neurons.x += torch.randint(
            -x_noise, x_noise, self.neurons.size, device=self.neurons.device
        )
        self.neurons.y += torch.randint(
            -y_noise, y_noise, self.neurons.size, device=self.neurons.device
        )
        self.neurons.z += torch.randint(
            -z_noise, z_noise, self.neurons.size, device=self.neurons.device
        )
        return self

    def rotate(self, axis, angle):
        """Rotate the `NeuronGroup` in space.

        Args:
            axis (int): The axis along which the rotation is made.
            angle (float): The angle to rotate in radians.

        Returns:
            NeuronDimension: The modified `NeuronDimension`.
        """
        rotation = rotation_matrix(axis, angle, self.neurons.def_dtype)
        self.neurons.x, self.neurons.y, self.neurons.z = torch.matmul(
            rotation, torch.stack([self.neurons.x, self.neurons.y, self.neurons.z])
        )
        return self

    def stretch_to_equal_size(self, target_neurons):
        """Stretch to `NeuronGroup` to match a target `NeuronGroup` dimensions.

        Args:
            target_neurons (NeuronGroup): The target `NeuronGroup`.
        """
        if hasattr(target_neurons, "width") and self.neurons.width > 0:
            x_stretch = target_neurons.width / self.neurons.width
            self.neurons.x *= x_stretch
        if hasattr(target_neurons, "height") and self.neurons.height > 0:
            y_stretch = target_neurons.height / self.neurons.height
            self.neurons.y *= y_stretch
        if hasattr(target_neurons, "depth") and self.neurons.depth > 0:
            z_stretch = target_neurons.depth / self.neurons.depth
            self.neurons.z *= z_stretch

    def initialize(self, neurons):
        super().initialize(neurons)

        self.width = self.parameter("width", 1, neurons)
        self.height = self.parameter("height", 1, neurons)
        self.depth = self.parameter("depth", 1, neurons)

        for pg in self.parameter("input_patterns", torch.tensor([]), neurons):
            dim = pg.size()
            if len(dim) > 0:
                self.height = max(self.height, dim[0])
            if len(dim) > 1:
                self.width = max(self.width, dim[1])
            if len(dim) > 2:
                self.depth = max(self.depth, dim[0])
                self.height = max(self.height, dim[1])
                self.width = max(self.width, dim[2])

        self.neurons = neurons

        neurons.shape = self
        neurons.width = self.width
        neurons.height = self.height
        neurons.depth = self.depth

        neurons.size = self.width * self.height * self.depth

        if neurons.network.index_neurons:
            self.set_position(self.width, self.height, self.depth)

            if self.parameter("centered", True, neurons):
                self.move(
                    -(self.width - 1) / 2, -(self.height - 1) / 2, -(self.depth - 1) / 2
                )



import torch

from pymonntorch.NetworkCore.Behavior import Behavior
from pymonntorch.utils import check_is_torch_tensor


class SynapticNormalization(Behavior):
    def __init__(self, *args, synapse_type="glutamate", norm_factor=1.0, **kwargs):
        super().__init__(
            *args, synapse_type=synapse_type, norm_factor=norm_factor, **kwargs
        )

    def initialize(self, neurons):
        super().initialize(neurons)

        self.synapse_type = self.parameter("synapse_type", "glutamate", neurons)

        neurons.require_synapses(self.synapse_type)

        self.norm_factor = check_is_torch_tensor(
            self.parameter("norm_factor", 1.0, neurons),
            device=neurons.device,
            dtype=neurons.def_dtype,
        )

        neurons.sum_w = neurons.vector(kwargs={"dtype": neurons.def_dtype})

    def forward(self, neurons):
        neurons.sum_w.zero_()

        for syn in neurons.afferent_synapses[self.synapse_type]:
            syn.dst.sum_w.add_(syn.w.sum(dim=1))

        neurons.sum_w.div_(self.norm_factor)

        for syn in neurons.afferent_synapses[self.synapse_type]:
            syn.w.T.div_(syn.dst.sum_w)



import copy

from pymonntorch.NetworkCore.Base import *


class SynapseGroup(NetworkObject):
    """This is the class to construct synapses between neuronal populations.

    Attributes:
        src (NeuronGroup): The pre-synaptic neuron group.
        dst (NeuronGroup): The post-synaptic neuron group.
        net (Network): The network the synapse group belongs to.
        tags (list): The tags of the synapse group.
        behavior (dict or list): The behaviors of the synapse group.
        enabled (bool): Whether the synapse is enabled for learning or not.
        group_weighting (float): The weighting of the synapse group.
    """

    def __init__(self, net, src=None, dst=None, tag=None, behavior=None):
        """This is the constructor of the SynapseGroup class.

        Args:
            src (NeuronGroup): The pre-synaptic neuron group.
            dst (NeuronGroup): The post-synaptic neuron group.
            net (Network): The network the synapse group belongs to.
            tag (str): The tag of the synapse group.
            behavior (dict or list): The behaviors of the synapse group.
        """
        if type(src) is str:
            src = net[src, 0]

        if type(dst) is str:
            dst = net[dst, 0]

        if tag is None and net is not None:
            tag = "SynapseGroup_" + str(len(net.SynapseGroups) + 1)

        super().__init__(net, tag, behavior, net.device)
        self.add_tag("syn")

        if net is not None:
            net.SynapseGroups.append(self)
            setattr(net, self.tags[0], self)

        self.src = src
        self.dst = dst
        self.enabled = True
        self.group_weighting = 1

        for ng in self.network.NeuronGroups:
            for tag in self.tags + ["All"]:
                if tag not in ng.afferent_synapses:
                    ng.afferent_synapses[tag] = []
                if tag not in ng.efferent_synapses:
                    ng.efferent_synapses[tag] = []

        if self.dst is not None:  
            self.connect_dst(self.dst)

        if self.src is not None:
            self.connect_src(self.src)

    def set_connection_tag(self):
        """Adds the connection tag
        """
        if self.src is not None and self.dst is not None:
            tag = self.src.tags[0] + " => " + self.dst.tags[0]
            if tag not in self.tags:
                self.add_tag(tag)
                for ng in self.network.NeuronGroups:
                    if tag not in ng.afferent_synapses:
                        ng.afferent_synapses[tag] = []
                    if tag not in ng.efferent_synapses:
                        ng.efferent_synapses[tag] = []

    def connect_src(self, src):
        """Connects the pre-synaptic neuron group to the synapse.
        
        Args:
            src (NeuronGroup): The pre-synaptic neuron group.
        """
        self.src = src
        # only add to NeuronGroup not to NeuronSubGroup
        if self.src.BaseNeuronGroup == self.src:
            for tag in self.tags + ["All"]:
                self.src.efferent_synapses[tag].append(self)
        self.set_connection_tag()

    def connect_dst(self, dst):
        """Connects the post-synaptic neuron group to the synapse.
        
        Args:
            dst (NeuronGroup): The post-synaptic neuron group.
        """
        self.dst = dst
        if self.dst.BaseNeuronGroup == self.dst:
            for tag in self.tags + ["All"]:
                self.dst.afferent_synapses[tag].append(self)
        self.set_connection_tag()

    def __repr__(self):
        result = "SynapseGroup" + str(self.tags)
        if self.src is not None and self.dst is not None:
            if self.network.transposed_synapse_matrix_mode:
                result = result + "(S" + str(self.src.size) + "xD" + str(self.dst.size)
            else:
                result = result + "(D" + str(self.dst.size) + "xS" + str(self.src.size)
        result = result + "){"

        for k in sorted(list(self.behavior.keys())):
            result += str(k) + ":" + str(self.behavior[k]) + ","
        return result + "}"

    def set_var(self, key, value):
        """Sets a variable of the synapse group.

        Args:
            key (str): The name of the variable.
            value (any): The value of the variable.

        Returns:
            SynapseGroup: The synapse group itself.
        """
        setattr(self, key, value)
        return self

    def matrix_dim(self):
        """Returns the dimension of the synapse matrix.

        For a synapse group between a source population of size n and a destination population of size m, the synapse matrix has the dimension m x n.

        Returns:
            tuple: The dimension of the synapse matrix.
        """
        if self.network.transposed_synapse_matrix_mode:
            return self.src.size, self.dst.size
        return self.dst.size, self.src.size

    def get_random_synapse_mat_fixed(self, min_number_of_synapses=0):
        """Returns a random synapse matrix with a fixed number of synapses per neuron.

        Args:
            min_number_of_synapses (int): The minimum number of synapses per neuron.

        Returns:
            torch.Tensor: The random synapse matrix.
        """
        dim = self.matrix_dim()
        result = torch.zeros(dim, device=self.device)
        if min_number_of_synapses != 0:
            for i in range(dim[0]):
                synapses = torch.randperm(dim[1], device=self.device)[
                    :min_number_of_synapses
                ]
                result[i, synapses] = torch.rand(len(synapses), device=self.device)
        return result

    def matrix(
        self,
        mode="zeros()",
        scale=None,
        density=None,
        only_enabled=True,
        clone_along_first_axis=False,
        plot=False,
        dtype=None,
    ):
        """Get a tensor with synapse group dimensionality.

        The tensor can be initialized in different modes. List of possible values for mode includes:
        - "random" or "rand" or "rnd" or "uniform": Uniformly distributed random numbers in range [0, 1).
        - "normal": Normally distributed random numbers with zero mean and unit variance.
        - "ones": Tensor filled with ones.
        - "zeros": Tensor filled with zeros.
        - A single number: Tensor filled with that number.
        - You can also use any function from torch package for this purpose. Note that you should **not** use `torch.` prefix.

        Args:
            mode (str): Mode to be used to initialize tensor.
            scale (float): Scale of the tensor. The default is None (i.e. No scaling is applied).
            density (float): Density of the tensor. The default is None (i.e. dense tensor).
            only_enabled (bool): Whether to only consider enabled synapses or not. The default is True.
            clone_along_first_axis (bool): Whether to clone the tensor along the first axis or not. The default is False.
            plot (bool): If true, the histogram of the tensor will be plotted. The default is False.
            dtype (str or type): Data type of the tensor. If None, `def_dtype` will be used.

        Returns:
            torch.Tensor: The initialized tensor.
        """
        result = self._get_mat(
            mode=mode,
            dim=(self.matrix_dim()),
            scale=scale,
            density=density,
            plot=plot,
            dtype=dtype,
        )

        if clone_along_first_axis:
            result = (
                torch.cat([result[0] for _ in range(self.matrix_dim()[0])])
                .reshape(self.matrix_dim()[0], *result[0].shape)
                .to(self.device)
            )

        if only_enabled:
            result *= self.enabled

        return result

    def get_synapse_group_size_factor(self, synapse_group, synapse_type):
        """Returns the size factor of a synapse group.

        Args:
            synapse_group (SynapseGroup): The synapse group.
            synapse_type (str): The type of the synapse.

        Returns:
            float: The size factor of the synapse group.
        """
        total_weighting = 0
        for s in synapse_group.dst.afferent_synapses[synapse_type]:
            total_weighting += s.group_weighting

        total = 0
        for s in synapse_group.dst.afferent_synapses[synapse_type]:
            total += s.src.size * s.src.group_weighting

        return (
            total_weighting
            / total
            * synapse_group.src.size
            * synapse_group.group_weighting
        )

    def get_distance_mat(self, radius, src_x=None, src_y=None, dst_x=None, dst_y=None):
        """Returns a distance matrix between source and destination neurons.

        Args:
            radius (float): The radius of the distance to be considered.
            src_x (torch.Tensor): The x coordinates of the source neurons. The default is None (i.e. the x coordinates of the source neurons will be used).
            src_y (torch.Tensor): The y coordinates of the source neurons. The default is None (i.e. the y coordinates of the source neurons will be used).
            dst_x (torch.Tensor): The x coordinates of the destination neurons. The default is None (i.e. the x coordinates of the destination neurons will be used).
            dst_y (torch.Tensor): The y coordinates of the destination neurons. The default is None (i.e. the y coordinates of the destination neurons will be used).

        Returns:
            torch.Tensor: The distance matrix.
        """
        if src_x is None:
            src_x = self.src.x
        if src_y is None:
            src_y = self.src.y
        if dst_x is None:
            dst_x = self.dst.x
        if dst_y is None:
            dst_y = self.dst.y

        result_syn_mat = torch.zeros((len(dst_x), len(src_x)), device=self.device)

        for d_n in range(len(dst_x)):
            dx = torch.abs(src_x - dst_x[d_n])
            dy = torch.abs(src_y - dst_y[d_n])

            dist = torch.sqrt(dx * dx + dy * dy)
            inv_dist = torch.clamp(radius - dist, 0.0, None)
            inv_dist /= torch.max(inv_dist)

            result_syn_mat[d_n] = inv_dist

        return result_syn_mat

    def get_ring_mat(
        self, radius, inner_exp, src_x=None, src_y=None, dst_x=None, dst_y=None
    ):
        """Returns a ring-shaped distance matrix between source and destination neurons.

        Args:
            radius (float): The radius of the ring.
            inner_exp (float): The exponent of the inner radius.
            src_x (torch.Tensor): The x coordinates of the source neurons. The default is None (i.e. the x coordinates of the source neurons will be used).
            src_y (torch.Tensor): The y coordinates of the source neurons. The default is None (i.e. the y coordinates of the source neurons will be used).
            dst_x (torch.Tensor): The x coordinates of the destination neurons. The default is None (i.e. the x coordinates of the destination neurons will be used).
            dst_y (torch.Tensor): The y coordinates of the destination neurons. The default is None (i.e. the y coordinates of the destination neurons will be used).

        Returns:
            torch.Tensor: The ring matrix.
        """
        dm = self.get_distance_mat(radius, src_x, src_y, dst_x, dst_y)
        ring = torch.clamp(dm - torch.pow(dm, inner_exp) * 1.5, 0.0, None)
        return ring / torch.max(ring)

    def get_max_receptive_field_size(self):
        """Returns the maximum receptive field size of the synapse group.

        Returns:
            tuple: The maximum receptive field size of the synapse group.
        """
        max_dx = 1
        max_dy = 1
        max_dz = 1

        for i in range(self.dst.size):
            if type(self.enabled) is torch.tensor:
                mask = self.enabled[i]
            else:
                mask = self.enabled

            if torch.sum(mask) > 0:
                x = self.dst.x[i]
                y = self.dst.y[i]
                z = self.dst.z[i]

                sx_v = self.src.x[mask]
                sy_v = self.src.y[mask]
                sz_v = self.src.z[mask]

                max_dx = torch.maximum(torch.max(torch.abs(x - sx_v)), max_dx)
                max_dy = torch.maximum(torch.max(torch.abs(y - sy_v)), max_dy)
                max_dz = torch.maximum(torch.max(torch.abs(z - sz_v)), max_dz)

        return max_dx, max_dy, max_dz

    def get_sub_synapse_group(self, src_mask, dst_mask):
        """Returns a sub synapse group between two neuronal subgroups.

        Args:
            src_mask (torch.Tensor): The mask of the source neurons.
            dst_mask (torch.Tensor): The mask of the destination neurons.

        Returns:
            SynapseGroup: The sub synapse group.
        """
        result = SynapseGroup(
            self.src.subGroup(src_mask),
            self.dst.subGroup(dst_mask),
            net=None,
            behavior={},
        )

        # partition enabled update
        if type(self.enabled) is torch.tensor:
            mat_mask = dst_mask[:, None] * src_mask[None, :]
            result.enabled = self.enabled[mat_mask].copy().reshape(result.matrix_dim())

        # copy al attributes
        sgd = self.__dict__
        for key in sgd:
            if key == "behavior":
                for k in self.behavior:
                    result.behavior[k] = copy.copy(self.behavior[k])
            elif key not in ["src", "dst", "enabled", "_mat_eval_dict"]:
                setattr(result, key, copy.copy(sgd[key]))

        return result



import torch

from pymonntorch.NetworkCore.Base import TaggableObject
from pymonntorch.utils import is_number


class Behavior(TaggableObject):
    """Base class for behaviors. All behaviors all `TaggableObject`s.

    Attributes:
        tag (str): Tag of the behavior.
        device (str): Device of the behavior. This is overwritten by object's device upon calling `initialize`.
        behavior_enabled (bool): Whether the behavior is enabled. The default is True.
        init_kwargs (dict): Dictionary of the keyword arguments passed to the constructor.
        used_attr_keys (list): List of the name of the attributes that have been used in the `initialize` method.
    """

    initialize_on_init = False
    initialize_last = False

    def __init__(self, *args, **kwargs):
        """Constructor of the `Behavior` class.

        Args:
            **kwargs: Keyword arguments passed to the constructor.
        """
        self.init_kwargs = kwargs
        for i, arg in enumerate(args):
            self.init_kwargs["arg_" + str(i)] = arg
        self.used_attr_keys = []
        self.behavior_enabled = self.parameter("behavior_enabled", True, None)
        super().__init__(
            tag=self.parameter("tag", None, None),
            device=self.parameter("device", None, None),
        )
        self.empty_iteration_function = self.is_empty_iteration_function()
        self.used_attr_keys = torch.nn.ParameterList(self.used_attr_keys)

    def initialize(self, object):
        """Sets the variables of the object. This method is called by the `Network` class when the object is added to the network.

        **Note:** All sub-classes of `Behavior` overriding this method should call the super method to ensure everything is placed on the correct device.

        Args:
            object (TaggableObject): Object possessing the behavior.
        """
        self.device = object.device
        return

    def forward(self, object):
        """Forward pass of the behavior. This method is called by the `Network` class per simulation iteration.

        Args:
            object (TaggableObject): Object possessing the behavior.
        """
        pass

    def __repr__(self):
        result = self.__class__.__name__ + "("
        for k in self.init_kwargs:
            result += str(k) + "=" + str(self.init_kwargs[k]) + ","
        result += ")"
        return result

    def evaluate_diversity_string(self, ds, object):
        """Evaluates the diversity string describing tensors of an object.

        Args:
            ds (str): Diversity string describing the tensors of the object.
            object (NetworkObject): The object possessing the behavior.

        Returns:
            torch.tensor: The resulting tensor.
        """
        if "same(" in ds and ds[-1] == ")":
            params = ds[5:-1].replace(" ", "").split(",")
            if len(params) == 2:
                return getattr(object[params[0], 0], params[1])

        plot = False
        if ";plot" in ds:
            ds = ds.replace(";plot", "")
            plot = True

        result = ds

        if "(" in ds and ")" in ds:  # is function
            if type(object).__name__ == "NeuronGroup":
                result = object.vector(ds)

            if type(object).__name__ == "SynapseGroup":
                result = object.matrix(ds)

        if plot:
            if type(result) == torch.tensor:
                import matplotlib.pyplot as plt

                plt.hist(result.to("cpu"), bins=30)
                plt.show()

        return result

    def set_parameters_as_variables(self, object):
        """Set the variables defined in the init of behavior as the variables of the object.

        Args:
            object (NetworkObject): The object possessing the behavior.
        """
        for key in self.init_kwargs:
            setattr(object, key, self.parameter(key, None, object=object))
            print("init", key)

    def check_unused_attrs(self):
        """Checks whether all attributes have been used in the `initialize` method."""
        for key in self.init_kwargs:
            if key not in self.used_attr_keys:
                print(
                    'Warning: "'
                    + key
                    + '" not used in initialize of '
                    + str(self)
                    + ' behavior! Make sure that "'
                    + key
                    + '" is spelled correctly and parameter('
                    + key
                    + ",...) is called in initialize. Valid attributes are: "
                    + ", ".join([f'"{param}"' for param in list(self.used_attr_keys)])
                    + "."
                )

    def parameter(
        self,
        key,
        default=None,
        object=None,
        do_not_diversify=False,
        search_other_behaviors=False,
        tensor=False,
        required=False,
    ):
        """Gets the value of an attribute.

        Args:
            key (str): Name of the attribute.
            default (any): Default value of the attribute.
            object (NetworkObject): The object possessing the behavior.
            do_not_diversify (bool): Whether to diversify the attribute. The default is False.
            search_other_behaviors (bool): Whether to search for the attribute in other behaviors of the object. The default is False.
            tensor (bool): Whether to make a tensor out of value. Suitable for list and numbers.
            required (bool): Whether the attribute is required. The default is False.

        Returns:
            any: The value of the attribute.
        """
        if required and self.init_kwargs.get(key, None) is None:
            print(
                "Warning:",
                key,
                "has to be specified for the behavior with a non None value to run properly.",
                self,
            )

        self.used_attr_keys.append(key)

        result = self.init_kwargs.get(key, default)
        result = default if result is None else result

        if (
            key not in self.init_kwargs
            and object is not None
            and search_other_behaviors
        ):
            for b in object.behaviors:
                if key in b.init_kwargs:
                    result = b.init_kwargs.get(key, result)

        if not do_not_diversify and type(result) is str and object is not None:
            result = self.evaluate_diversity_string(result, object)

        if type(result) is str and default is not None:
            if "%" in result and is_number(result.replace("%", "")):
                result = str(float(result.replace("%", "")) / 100.0)

            result = type(default)(result)

        if tensor and result is not None:
            if object is None:
                raise RuntimeError(
                    f'To turn parameter value of key "{key}" to a tensor, object should not be None.'
                )
            result = torch.tensor(result, device=object.device)
            if result.is_floating_point():
                result = result.to(dtype=object.def_dtype)

        return result

    def is_empty_iteration_function(self):
        """Checks whether a function does anything or not.

        used to stop calling behaviors with empty forward method.
        """
        f = self.forward

        # Returns true if f is an empty function.
        def empty_func():
            pass

        def empty_func_with_docstring():
            """Empty function with docstring."""
            pass

        def constants(f):
            # Return a tuple containing all the constants of a function without: * docstring
            return tuple(x for x in f.__code__.co_consts if x != f.__doc__)

        return (
            f.__code__.co_code == empty_func.__code__.co_code
            and constants(f) == constants(empty_func)
        ) or (
            f.__code__.co_code == empty_func_with_docstring.__code__.co_code
            and constants(f) == constants(empty_func_with_docstring)
        )



import torch

from pymonntorch.NetworkCore.Behavior import Behavior
from pymonntorch.utils import check_is_torch_tensor


class InstantHomeostasis(Behavior):
    def __init__(
        self,
        *args,
        min_th=None,
        max_th=None,
        threshold=None,
        gap_percent=None,
        distance_sensitive=True,
        inc=1.0,
        dec=1.0,
        adj_strength=1.0,
        adjustment_param=None,
        measurement_param=None,
        measurement_min=None,
        measurement_max=None,
        target_clip_min=None,
        target_clip_max=None,
        **kwargs
    ):
        super().__init__(
            *args,
            min_th=min_th,
            max_th=max_th,
            threshold=threshold,
            gap_percent=gap_percent,
            distance_sensitive=distance_sensitive,
            inc=inc,
            dec=dec,
            adj_strength=adj_strength,
            adjustment_param=adjustment_param,
            measurement_param=measurement_param,
            measurement_min=measurement_min,
            measurement_max=measurement_max,
            target_clip_min=target_clip_min,
            target_clip_max=target_clip_max,
            **kwargs
        )

    def set_threshold_boundaries(self, th, gap_percent):
        self.th = th

        if th is not None:
            has_min = False
            has_max = False
            if self.min_th is None:
                self.min_th = th
                has_min = True
            if self.max_th is None:
                self.max_th = th
                has_max = True

            if gap_percent is not None:
                if has_min:
                    self.min_th -= self.min_th / 100 * gap_percent
                    check_is_torch_tensor(
                        self.min_th, device=self.device, dtype=self.dtype
                    )
                if has_max:
                    self.max_th += self.max_th / 100 * gap_percent
                    check_is_torch_tensor(
                        self.max_th, device=self.device, dtype=self.dtype
                    )

    def get_measurement_param(self, object):
        if self.compiled_mp is None:
            self.compiled_mp = compile(self.measurement_param, "<string>", "eval")

        result = check_is_torch_tensor(
            eval(self.compiled_mp), device=object.device, dtype=self.dtype
        )

        result.clamp_(self.measurement_min, self.measurement_max)
        return result

    def get_target_adjustment(self, measure):
        greater = (measure > self.max_th) * (-self.dec)
        smaller = (measure < self.min_th) * self.inc

        if self.distance_sensitive:
            greater *= measure - self.max_th
            smaller *= self.min_th - measure

        return (greater + smaller) * self.adj_strength

    def adjust_target(self, object, adj):
        val = check_is_torch_tensor(
            getattr(object, self.adjustment_param),
            device=object.device,
            dtype=self.dtype,
        )
        val.add_(adj)

        val.clamp_(self.target_clip_min, self.target_clip_max)

        setattr(object, self.adjustment_param, val)

    def initialize(self, object):
        super().initialize(object)

        self.dtype = object.def_dtype

        self.compiled_mp = None

        self.min_th = self.parameter(
            "min_th", None, object
        )  # minimum threshold for measurement param
        self.max_th = self.parameter(
            "max_th", None, object
        )  # maximum threshold for measurement param

        self.set_threshold_boundaries(
            self.parameter(
                "threshold", None, object=object
            ),  # threshold for measurement param (min=max=th)
            self.parameter(
                "gap_percent", None, object=object
            ),  # min max gap is created via a percentage from th (additional param for th)
        )

        self.distance_sensitive = self.parameter(
            "distance_sensitive", True, object
        )  # stronger adjustment when value is further away from optimum

        self.inc = check_is_torch_tensor(
            self.parameter("inc", 1.0, object),  # increase factor
            device=object.device,
            dtype=self.dtype,
        )
        self.dec = check_is_torch_tensor(
            self.parameter("dec", 1.0, object),  # decrease factor
            device=object.device,
            dtype=self.dtype,
        )

        self.adj_strength = check_is_torch_tensor(
            self.parameter("adj_strength", 1.0, object),  # change factor
            device=object.device,
            dtype=self.dtype,
        )

        self.adjustment_param = self.parameter(
            "adjustment_param", None, object
        )  # name of object target attribute

        self.measurement_param = self.parameter(
            "measurement_param", None, object
        )  # name of parameter to be measured

        self.measurement_min = check_is_torch_tensor(
            self.parameter(
                "measurement_min", None, object
            ),  # minimum value which can be measured (below=0)
            device=object.device,
            dtype=self.dtype,
        )
        self.measurement_max = check_is_torch_tensor(
            self.parameter(
                "measurement_max", None, object
            ),  # maximum value which can be measured (above=max)
            device=object.device,
            dtype=self.dtype,
        )

        self.target_clip_min = check_is_torch_tensor(
            self.parameter("target_clip_min", None, object),  # target clip min
            device=object.device,
            dtype=self.dtype,
        )
        self.target_clip_max = check_is_torch_tensor(
            self.parameter("target_clip_max", None, object),  # target clip max
            device=object.device,
            dtype=self.dtype,
        )

    def forward(self, object):
        measure = self.get_measurement_param(object)
        self.adj = self.get_target_adjustment(measure)
        self.adjust_target(object, self.adj)


class TimeIntegratedHomeostasis(InstantHomeostasis):
    def __init__(
        self,
        *args,
        min_th=None,
        max_th=None,
        threshold=None,
        gap_percent=None,
        distance_sensitive=True,
        inc=1.0,
        dec=1.0,
        adj_strength=1.0,
        adjustment_param=None,
        measurement_param=None,
        measurement_min=None,
        measurement_max=None,
        target_clip_min=None,
        target_clip_max=None,
        integration_length=1,
        init_avg=None,
        **kwargs
    ):
        super().__init__(
            *args,
            min_th=min_th,
            max_th=max_th,
            threshold=threshold,
            gap_percent=gap_percent,
            distance_sensitive=distance_sensitive,
            inc=inc,
            dec=dec,
            adj_strength=adj_strength,
            adjustment_param=adjustment_param,
            measurement_param=measurement_param,
            measurement_min=measurement_min,
            measurement_max=measurement_max,
            target_clip_min=target_clip_min,
            target_clip_max=target_clip_max,
            integration_length=integration_length,
            init_avg=init_avg,
            **kwargs
        )

    def get_measurement_param(self, object):
        val = super().get_measurement_param(object)

        self.average = (self.average * self.integration_length + val) / (
            self.integration_length + 1
        )
        return self.average

    def initialize(self, object):
        super().initialize(object)

        self.integration_length = check_is_torch_tensor(
            self.parameter("integration_length", 1, object),
            device=object.device,
            dtype=object.def_dtype,
        )
        self.average = check_is_torch_tensor(
            self.parameter("init_avg", (self.min_th + self.max_th) / 2, object),
            device=object.device,
            dtype=object.def_dtype,
        )



import inspect

from pymonntorch.NetworkCore.Base import *

# This class can be used to add tag-searchable functions to the neurongroups, synapsegroups and the network object.
# It has a main execute function which can be called with module(...) or module.exec(...)
# Other "normal" functions can be added as well.
# Via the add_tag function, the modules can be categorized into groups


class AnalysisModule(TaggableObject):
    """This class can be used to add tag-searchable functions to the neurongroups, synapsegroups and the network object.

    It has a main execute function which can be called with module(...) or module.exec(...).
    Other "normal" functions can be added as well. Via the add_tag function, the modules can be categorized into groups.

    Attributes:
        parent (NeuronGroup, SynapseGroup or Network): The parent object.
        init_kwargs (dict): The arguments passed to the constructor.
        used_attr_keys (list): The list of the used attributes.
        execution_arguments (dict): The arguments of the execute function.
        result_storage (dict): The dictionary of argument results after the execute function.
        current_result (any): The result of the last execute function call.
        save_results (bool): If the results should be saved.
        update_notifier_functions (list): The list of the update notifier functions.
        progress_update_function (function): The function to update the progress.
    """

    def __init__(self, parent=None, **kwargs):
        self.init_kwargs = kwargs
        super().__init__(tag=self.parameter("tag", None))

        self.used_attr_keys = []

        self.add_tag(self.__class__.__name__)

        self.execution_arguments = self._get_default_args_(self.execute)

        self.result_storage = {}  # key=arguments value=result from run
        self.current_result = None
        self.save_results = True

        self.update_notifier_functions = []

        self.progress_update_function = None

        if parent is not None:
            self._attach_and_initialize_(parent)

    def add_progress_update_function(self, function):
        """Adds a function to update the progress of the module.

        Args:
            function (function): The function to update the progress.
        """
        self.progress_update_function = function

    def update_progress(self, percent):
        """Updates the progress of the module.

        Args:
            percent (int): The progress in percent.
        """
        if self.progress_update_function is not None:
            self.progress_update_function(percent)

    def _attach_and_initialize_(self, parent):
        """Attaches the module to the parent object and initializes it.

        Args:
            parent (NeuronGroup, SynapseGroup or Network): The parent object.
        """
        self.parent = parent
        parent.analysis_modules.append(self)
        setattr(parent, self.__class__.__name__, self)
        self.initialize(parent)

    def initialize(self, object):
        """This function is called when the module is attached to the parent object. It should be overridden.

        - access arguments via self.parameter(key, default)
        - add tag via self.add_tag(tag)
        - add execution arguments via self.add_execution_argument(...)
        `execute` does not have to be used.

        Args:
            object (NeuronGroup, SynapseGroup or Network): The parent object.
        """
        return

    def execute(self, object, **kwargs):
        """Executes the functions of the module with the given arguments. It should be overridden.

        Note: Do not call this function directly. Use the instance(...) instead of instance.execute(...).

        Args:
            object (NeuronGroup, SynapseGroup or Network): The parent object.
            **kwargs: The arguments of the function.
        """
        return

    def is_executable(self):
        return type(self).execute != AnalysisModule.execute

    def parameter(self, key, default):
        """Returns the value of the given key from the init arguments. If the key is not present, the default value is returned.

        Args:
            key (str): The name of the argument.
            default (any): The default value.

        Returns:
            any: The value of the argument.
        """
        if key in self.init_kwargs:
            return self.init_kwargs[key]
        else:
            return default

    def _update_notification_(self, key=None):
        for function in self.update_notifier_functions:
            function(key)

    def remove_update_notifier(self, function):
        """Removes the given function from the update notifier functions.

        Args:
            function (function): The function to remove.
        """
        if function in self.update_notifier_functions:
            self.update_notifier_functions.remove(function)

    def set_update_notifier(self, function):
        """Adds the given function ro the list of update notifier functions.

        Args:
            function (function): The function to set.
        """
        self.update_notifier_functions.append(function)

    def __call__(self, **kwargs):
        self.update_progress(0)
        self.current_key = self.generate_current_key(kwargs)
        return self.save_result(self.current_key, self.execute(self.parent, **kwargs))

    def exec(self, **kwargs):
        """Executes the module with the given arguments.

        Args:
            **kwargs: The arguments of the function.

        Returns:
            any: The result of the function.
        """
        self.update_progress(0)
        self.current_key = self.generate_current_key(kwargs)
        return self.save_result(self.current_key, self.execute(self.parent, **kwargs))

    def _get_name(self):
        return self.__class__.__name__

    def _get_base_name(self):
        args = [str(v) for v in self.init_kwargs.values()]
        return self._get_name() + "(" + ", ".join(args) + ")"

    def generate_current_key(self, args_key, add_args=True):
        """Generates the key for the current execution.

        Args:
            args_key (list): The arguments of the function.
            add_args (bool): If the arguments should be added to the key.
        """
        key = self._get_base_name_()
        if len(args_key) > 0 and add_args:
            key += " " + str(args_key)
        return key

    def save_result(self, key, result):
        """Saves the result of the execution.

        Args:
            key (str): The key of the execution.
            result (any): The result of the execution.

        Returns:
            any: The result of the execution.
        """
        self.update_progress(100)
        if self.save_results and result is not None:
            self.current_result = result
            self.result_storage[key] = result
        self._update_notification_(key)
        return result

    def last_call_result(self):
        """Returns the result of the last execution."""
        return self.current_result

    def get_results(self):
        """Returns the results of all executions."""
        return self.result_storage

    def remove_result(self, key):
        """Removes the result of the given key from the dictionary of results.

        Args:
            key (str): The key of the execution to remove.
        """
        if key in self.result_storage:
            return self.result_storage.pop(key)
        else:
            print("cannot remove result", key, "not found.")
        self._update_notification_()

    def clear_results(self):
        """Removes all results."""
        self.result_storage = {}
        self._update_notification_()

    def _get_default_args_(
        self, func, exclude=["self", "args", "kwargs"], exclude_first=True
    ):
        result = {}
        signature = inspect.signature(func)
        i = 0
        for k, v in signature.parameters.items():
            if i > 0 or not exclude_first:
                if k not in exclude:
                    if v.default is not inspect.Parameter.empty:
                        result[k] = v.default
                    else:
                        result[k] = ""
            i += 1
        return result



#!/usr/bin/env python

"""Tests for `pymonntorch` package."""

import pytest

# from click.testing import CliRunner

# from pymonntorch import cli
import pymonntorch


def test_network_cpu_annotated():
    from pymonntorch import (
        Network,
        Behavior,
        NeuronGroup,
        SynapseGroup,
        Recorder,
        EventRecorder,
    )

    class BasicBehavior(Behavior):
        def initialize(self, neurons):
            super().initialize(neurons)
            neurons.voltage = neurons.vector(mode="zeros")
            self.threshold = 1.0

        def forward(self, neurons):
            firing = neurons.voltage >= self.threshold
            neurons.spike = firing.byte()
            neurons.voltage[firing] = 0.0  # reset

            neurons.voltage *= 0.9  # voltage decay
            neurons.voltage += neurons.vector(mode="uniform", density=0.1)

    class InputBehavior(Behavior):
        def initialize(self, neurons):
            super().initialize(neurons)
            for synapse in neurons.afferent_synapses["GLUTAMATE"]:
                synapse.W = synapse.matrix("uniform", density=0.1)
                synapse.enabled = synapse.W > 0

        def forward(self, neurons):
            for synapse in neurons.afferent_synapses["GLUTAMATE"]:
                neurons.voltage += (
                    synapse.W @ synapse.src.spike.float() / synapse.src.size * 10
                )

    net = Network()
    ng = NeuronGroup(
        net=net,
        size=100,
        behavior={
            1: BasicBehavior(),
            2: InputBehavior(),
            9: Recorder(["n.voltage", "torch.mean(n.voltage)"], auto_annotate=False),
            10: EventRecorder(["n.spike"]),
        },
    )
    SynapseGroup(ng, ng, net, tag="GLUTAMATE")
    net.initialize()
    net.simulate_iterations(1000)

    assert net.device == "cpu"
    assert ng.device == "cpu"

    for _, b in ng.behavior.items():
        assert b.device == "cpu"


def test_network_cuda_annotated():
    from pymonntorch import (
        Network,
        Behavior,
        NeuronGroup,
        SynapseGroup,
        Recorder,
        EventRecorder,
    )

    class BasicBehavior(Behavior):
        def initialize(self, neurons):
            super().initialize(neurons)
            neurons.voltage = neurons.vector(mode="zeros")
            self.threshold = 1.0

        def forward(self, neurons):
            firing = neurons.voltage >= self.threshold
            neurons.spike = firing.byte()
            neurons.voltage[firing] = 0.0  # reset

            neurons.voltage *= 0.9  # voltage decay
            neurons.voltage += neurons.vector(mode="uniform", density=0.1)

    class InputBehavior(Behavior):
        def initialize(self, neurons):
            super().initialize(neurons)
            for synapse in neurons.afferent_synapses["GLUTAMATE"]:
                synapse.W = synapse.matrix("uniform", density=0.1)
                synapse.enabled = synapse.W > 0

        def forward(self, neurons):
            for synapse in neurons.afferent_synapses["GLUTAMATE"]:
                neurons.voltage += (
                    synapse.W @ synapse.src.spike.float() / synapse.src.size * 10
                )

    net = Network(device="cuda")
    ng = NeuronGroup(
        net=net,
        size=100,
        behavior={
            1: BasicBehavior(),
            2: InputBehavior(),
            9: Recorder(["n.voltage", "torch.mean(n.voltage)"], auto_annotate=False),
            10: EventRecorder(["n.spike"]),
        },
    )
    SynapseGroup(ng, ng, net, tag="GLUTAMATE")
    net.initialize()
    net.simulate_iterations(1000)

    assert net.device == "cuda"
    assert ng.device == "cuda"

    for _, b in ng.behavior.items():
        assert b.device == "cuda"


# def test_command_line_interface():
#     """Test the CLI."""
#     runner = CliRunner()
#     result = runner.invoke(cli.main)
#     assert result.exit_code == 0
#     assert 'pymonntorch.cli.main' in result.output
#     help_result = runner.invoke(cli.main, ['--help'])
#     assert help_result.exit_code == 0
#     assert '--help  Show this message and exit.' in help_result.output



import numpy as np
import torch

from pymonntorch.NetworkCore.Base import *
from pymonntorch.NetworkCore.Behavior import *


class NeuronGroup(NetworkObject):
    """This is the class to construct a neuronal population.

    Attributes:
        size (int): The number of neurons in the population.
        behavior (list or dict): The behaviors of the population.
        net (Network): The network the population belongs to.
        tags (str): The tags of the population.
        BaseNeuronGroup (NeuronGroup): The base `NeuronGroup` the population belongs to.
        afferent_synapses (dict): The afferent synapses of the population.
        efferent_synapses (dict): The efferent synapses of the population.
        mask (bool): Whether to define a mask for the population (Used for nested populations).
        learning (bool): Whether to enable learning for the population.
        recording (bool): Whether to enable recording for the population.
        id (torch.Tensor): The integer id of the population.
    """

    def __init__(self, size, net, behavior=None, tag=None):
        """Initialize the neuronal population.

        Args:
            size (int or Behavior): The size or dimension of the population.
            behavior (list or dict): The behaviors of the population.
            net (Network): The network the population belongs to.
            tag (str): The tag of the population.
        """
        if tag is None and net is not None:
            tag = "NeuronGroup_" + str(len(net.NeuronGroups) + 1)

        if isinstance(size, Behavior):
            if type(behavior) is dict:
                if 0 in behavior:
                    print(
                        "warning: 0 index behavior will be overwritten by size behavior"
                    )
                behavior[0] = size
            if type(behavior) is list:
                behavior.insert(0, size)
            size = -1  # will be overwritten by size-behavior

        self.size = size

        super().__init__(net, tag, behavior, net.device)
        self.add_tag("ng")

        self.BaseNeuronGroup = self  # used for subgroup reconstruction

        if net is not None:
            net.NeuronGroups.append(self)
            setattr(net, self.tags[0], self)

        self.afferent_synapses = {}  # set by Network
        self.efferent_synapses = {}

        self.mask = True

        if net.index_neurons:
            self.id = torch.arange(self.size, device=self.device)

    def synapses(self, mode, tag="All"):
        """Get synapses connected to the NeuronGroup.

        Args:
            mode (str or bool): Whether to return efferent or afferent synapses connected to the NeuronGroup.
                                Values indicating afferent: ` "afferent", "dendrite", "pre", "preSynaptic", 0, False `
                                Values indicating efferent: ` "efferent", "axon", "post", "postSynaptic", 1 , True`
            tag (str, optional): Filters Synapses to have provided tag . Defaults to "All".

        Returns:
            List[SynapseGroup]: The list containing SynapseGroups
        """
        if mode in ["afferent", "dendrite", "pre", "preSynaptic", 0]:
            return self.afferent_synapses[tag]
        if mode in ["efferent", "axon", "post", "postSynaptic", 1]:
            return self.efferent_synapses[tag]

    def require_synapses(self, name, afferent=True, efferent=True, warning=True):
        """Require the existence of synapses.

        Args:
            name (str): The name of the synapse.
            afferent (bool): Whether to require afferent synapses.
            efferent (bool): Whether to require efferent synapses.
            warning (bool): Whether to print a warning if the synapse does not exist.
        """
        if afferent and not name in self.afferent_synapses:
            if warning:
                print("warning: no afferent {} synapses found".format(name))
            self.afferent_synapses[name] = []

        if efferent and not name in self.efferent_synapses:
            if warning:
                print("warning: no efferent {} synapses found".format(name))
            self.efferent_synapses[name] = []

    def vector(self, mode="zeros()", scale=None, density=None, plot=False, dtype=None):
        """Get a tensor with population's dimensionality.

        The tensor can be initialized in different modes. List of possible values for mode includes:
        - "random" or "rand" or "rnd" or "uniform": Uniformly distributed random numbers in range [0, 1).
        - "normal": Normally distributed random numbers with zero mean and unit variance.
        - "ones": Tensor filled with ones.
        - "zeros": Tensor filled with zeros.
        - A single number: Tensor filled with that number.
        - You can also use any function from torch package for this purpose. Note that you should **not** use `torch.` prefix.

        Args:
            mode (str): Mode to be used to initialize tensor.
            scale (float): Scale of the tensor. The default is None (i.e. No scaling is applied).
            density (float): Density of the tensor. The default is None (i.e. dense tensor).
            plot (bool): If true, the histogram of the tensor will be plotted. The default is False.
            dtype (str or type): Data type of the tensor. If None, `def_dtype` will be used.

        Returns:
            torch.Tensor: The initialized tensor."""
        return self._get_mat(
            mode=mode,
            dim=(self.size,),
            scale=scale,
            density=density,
            plot=plot,
            dtype=dtype,
        )

    def vector_buffer(self, buffer_size, **kwargs):
        """Get a buffer for the population's dimensionality.

        Args:
            buffer_size (int): The size of the buffer.
            **dtype (torch.dtype, optional): The desired data type of returned tensor.

        Returns:
            torch.Tensor: The buffer.
        """
        kwargs.setdefault("dtype", self.def_dtype)
        return self.get_buffer_mat((self.size,), buffer_size, **kwargs)

    def get_combined_synapse_shape(self, Synapse_ID):
        """Get the population size along with the number of afferent synapses.

        Args:
            Synapse_ID (str): The ID of the synapse by which it is registered in list of afferent synapses.

        Returns:
            tuple: The combined shape.
        """
        source_num = 0
        for syn in self.afferent_synapses[Synapse_ID]:
            _, s = syn.matrix_dim()
            source_num += s
        return self.size, source_num

    def __repr__(self):
        result = "NeuronGroup" + str(self.tags) + "(" + str(self.size) + "){"
        for k in sorted(list(self.behavior.keys())):
            result += str(k) + ":" + str(self.behavior[k])
        return result + "}"

    def subGroup(self, mask=None):
        """Get a NeuronSubGroup object from the population.

        Args:
            mask (bool): The mask condition indicating which neurons to be included in the subgroup.

        Returns:
            NeuronSubGroup: The subgroup.
        """
        return NeuronSubGroup(self, mask)

    def group_without_subGroup(self):
        """Get the NeuronGroup object itself."""
        return self

    def get_masked_dict(self, dict_name, key):
        """Get value of a key in a specific dictionary attribute of the population.

        Args:
            dict_name (str): Name of the dictionary.
            key (int or str): the key to retrieve from the dictionary.

        Returns:
            any: The value.
        """
        return getattr(self, dict_name)[key]

    def connected_NG_param_list(
        self,
        param_name,
        syn_tag="All",
        afferent_NGs=False,
        efferent_NGs=False,
        same_NG=False,
        search_behaviors=False,
    ):
        """Get a list of parameters of connected neuron groups.

        Args:
            param_name (str): The name of the parameter.
            syn_tag (str): The tag of the synapse. The default is "All".
            afferent_NGs (bool): Whether to include afferent neuron groups. The default is False.
            efferent_NGs (bool): Whether to include efferent neuron groups. The default is False.
            same_NG (bool): Whether to include the connections within the same neuron group. The default is False.
            search_behaviors (bool): Whether to search the dictionary of behaviors for the parameter. The default is False.

        Returns:
            list: The list of parameters.
        """
        result = []

        def search_NG(NG):
            if hasattr(NG, param_name):
                attr = getattr(NG, param_name)
                if callable(attr):
                    result.append(attr(NG))
                else:
                    result.append(attr)
            if search_behaviors:
                for key, behavior in NG.behavior.items():
                    if hasattr(behavior, param_name):
                        attr = getattr(behavior, param_name)
                        if callable(attr):
                            result.append(attr(NG))
                        else:
                            result.append(attr)

        if same_NG:
            search_NG(self)

        if efferent_NGs:
            for syn in self.efferent_synapses[syn_tag]:
                search_NG(syn.dst)

        if afferent_NGs:
            for syn in self.afferent_synapses[syn_tag]:
                search_NG(syn.src)

        return result

    def partition(self, block_size=7):
        """Get a partitioned population.

        Args:
            block_size (int): The size of each block. The default is 7.

        Returns:
            list of NeuronGroup or NeuronSubGroup: The list of partitions.
        """
        w = block_size
        h = block_size
        d = block_size
        split_size = [np.maximum(w, 1), np.maximum(h, 1), np.maximum(d, 1)]
        if split_size[0] < 2 and split_size[1] < 2 and split_size[2] < 2:
            return [self]
        else:
            return self.split_grid_into_sub_group_blocks(split_size)

    def partition_masks(self, steps=[1, 1, 1]):
        """Get a mask tensor for partitioning the population.

        Args:
            steps (list of int): The number of steps in each dimension. The default is [1, 1, 1].

        Returns:
            torch.Tensor: The mask tensor.
        """
        dst_min = [np.min(p) for p in [self.x, self.y, self.z]]
        dst_max = [np.max(p) for p in [self.x, self.y, self.z]]

        def get_start_end(step, dim):
            start = dst_min[dim] + (dst_max[dim] - dst_min[dim]) / steps[dim] * step
            end = dst_min[dim] + (dst_max[dim] - dst_min[dim]) / steps[dim] * (step + 1)
            return start, end

        results = []

        masks = []
        for w_step in range(steps[0]):  # x_steps
            dst_x_start, dst_x_end = get_start_end(w_step, 0)
            for h_step in range(steps[1]):  # y_steps
                dst_y_start, dst_y_end = get_start_end(h_step, 1)
                for d_step in range(steps[2]):  # z_steps
                    dst_z_start, dst_z_end = get_start_end(d_step, 2)

                    sub_group_mask = (
                        (self.x >= dst_x_start)
                        * (self.x <= dst_x_end)
                        * (self.y >= dst_y_start)
                        * (self.y <= dst_y_end)
                        * (self.z >= dst_z_start)
                        * (self.z <= dst_z_end)
                    )

                    # remove redundancies
                    for old_dst_mask in masks:
                        sub_group_mask[old_dst_mask] *= False
                    masks.append(sub_group_mask)

                    results.append(sub_group_mask)

        return torch.tensor(results).to(self.device)

    def split_grid_into_sub_group_blocks(self, steps=[1, 1, 1]):
        """Split the population into partitioned subgroups.

        Returns:
            list of NeuronGroup or NeuronSubGroup: The list of partitions.
        """
        return [self.subGroup(mask) for mask in self.partition_masks(steps)]

    def get_subgroup_receptive_field_mask(self, subgroup, xyz_rf=[1, 1, 1]):
        """Get the receptive field mask of a neuron subgroup.

        Args:
            subgroup (NeuronSubGroup or NeuronGroup): The neuron subgroup.
            xyz_rf (list of int): The receptive field size in each dimension. The default is [1, 1, 1].

        Returns:
            torch.Tensor: The receptive field mask.
        """
        rf_x, rf_y, rf_z = xyz_rf

        src_x_start = np.min(subgroup.x) - rf_x
        src_x_end = np.max(subgroup.x) + rf_x

        src_y_start = np.min(subgroup.y) - rf_y
        src_y_end = np.max(subgroup.y) + rf_y

        src_z_start = np.min(subgroup.z) - rf_z
        src_z_end = np.max(subgroup.z) + rf_z

        mask = (
            (self.x >= src_x_start)
            * (self.x <= src_x_end)
            * (self.y >= src_y_start)
            * (self.y <= src_y_end)
            * (self.z >= src_z_start)
            * (self.z <= src_z_end)
        )

        return mask

    def mask_var(self, var):
        """Mask a variable.

        Args:
            var (torch.Tensor): The variable.

        Returns:
            torch.Tensor: The masked variable.
        """
        return var


class NeuronSubGroup:
    def __init__(self, BaseNeuronGroup, mask):
        self.cache = {}
        self.key_id_cache = {}
        self.BaseNeuronGroup = BaseNeuronGroup
        self.mask = mask
        self.id_mask = torch.from_numpy(np.where(mask)[0]).to(BaseNeuronGroup.device)

    def mask_var(self, var):
        if var.shape[0] != self.mask.shape[0]:
            return var[:, self.mask]
        else:
            return var[self.mask]

    def __getattr__(self, attr_name):
        if attr_name in ["BaseNeuronGroup", "mask", "cache", "key_id_cache", "id_mask"]:
            return super().__getattr__(attr_name)  # setattr

        if attr_name == "size":
            return torch.sum(self.mask)

        attr = getattr(self.BaseNeuronGroup, attr_name)
        if type(attr) == torch.tensor:
            if attr.shape[0] != self.mask.shape[0]:
                return attr[:, self.mask]
            else:
                return attr[self.mask]
        else:
            return attr

    def __setattr__(self, attr_name, value):
        if attr_name in ["BaseNeuronGroup", "mask", "cache", "key_id_cache", "id_mask"]:
            super().__setattr__(attr_name, value)
            return

        attr = getattr(self.BaseNeuronGroup, attr_name)
        if type(attr) == torch.tensor:
            if attr.shape[0] != self.mask.shape[0]:
                attr[:, self.mask] = value
            else:
                attr[self.mask] = value
        else:
            setattr(self.BaseNeuronGroup, attr, value)

    def group_without_subGroup(self):
        return self.BaseNeuronGroup



from pymonntorch.NetworkCore.Behavior import Behavior

from sympy.physics.units import convert_to, second, seconds


class Clock(Behavior):
    def __init__(self, *args, step="1*ms", **kwargs):
        super().__init__(*args, step=step, **kwargs)

    def initialize(self, neuron_or_network):
        super().initialize(neuron_or_network)

        self.add_tag("Clock")

        neuron_or_network.clock_step_size = float(
            convert_to(eval(self.parameter("step", "1*ms")), seconds) / second
        )  # in ms (*ms)
        self.clock_step_size = neuron_or_network.clock_step_size
        print(neuron_or_network.clock_step_size)
        neuron_or_network.t = 0.0

    def forward(self, neuron_or_network):
        neuron_or_network.t += neuron_or_network.clock_step_size

    def time_to_iterations(self, time_str):
        iterations = int(
            convert_to(eval(time_str + "/seconds"), seconds) / self.clock_step_size
        )
        return iterations



import time

import torch

from pymonntorch.NetworkCore.Base import *
from pymonntorch.NetworkCore.Behavior import Behavior
from pymonntorch.NetworkCore.SynapseGroup import *


class Network(NetworkObject):
    """This is the class to construct a neural network.

    This is the placeholder of all neural network components to be simulated.
    All objects will receive an instance of this class.

    Attributes:
        NeuronGroups (list): List of all NeuronGroups in the network.
        SynapseGroups (list): List of all SynapseGroups in the network.
        behavior (list or dict): List of all network-specific behaviors.
        def_dtype (type): Floating point precision of tensors. Defaults to `torch.float32`.
        device (string): The device to allocate tensors on. Defaults to `'cpu'`.
        transposed_synapse_matrix_mode (bool): If `True`, in the matrix created by synapse, each row corresponds to a neuron from the source neuron group. Defaults to `False`.
        index_neurons (bool): If True, the `id` attribute for neuron groups refers to neuron indices. Defaults to `True`.
    """

    def __init__(
        self,
        tag=None,
        behavior=None,
        dtype=torch.float32,
        device="cpu",
        synapse_mode="SxD",
        index=True,
    ):
        """Initialize the network.

        Args:
            tag (str): Tag to add to the network. It can also be a comma-separated string of multiple tags.
            behavior (list or dict): List or dictionary of behaviors. If a dictionary is used, the keys must be integers.
            dtype (torch.dtype): Floating point precision of tensors.
            synapse_mode (string): If `SxD`, rows of matrix created by a synapse referes to the source neuron group and columns referes to the Destination neurpn group. Possible values `DxS` and `SxD`.
            device (string): The device to allocate tensors' data on. Defaults to `'cpu'`.
            index (bool): If True, create an indexing tensor as `id` attribute for each neuron group.
        """
        self.device = device
        self._def_dtype = dtype
        self.index_neurons = index
        if synapse_mode not in ["SxD", "DxS"]:
            print(
                f"""warning: synapse_mode should be either "SxD" or "DxS". using "DxS"."""
            )
        self.transposed_synapse_matrix_mode = (
            False if synapse_mode == "DxS" else synapse_mode == "SxD"
        )

        self.NeuronGroups = []
        self.SynapseGroups = []
        self.Structures = []

        self._iteration = 0

        self.sorted_behavior_execution_list = (
            []
        )  # stores (key, beh_parent, behavior) triplets

        super().__init__(self, tag, behavior, device=self.device)

    def fill_substructures(self):
        self.sub_structures = []
        for struc in self.Structures:
            if struc.parent_structure == self:
                self.sub_structures.append(struc)
    
    def set_behaviors(self, tag, enabled):
        """Set behaviors of specific tag to be enabled or disabled.

        Args:
            tag (str): Tag of behaviors to be enabled or disabled.
            enabled (bool): If true, behaviors will be enabled. If false, behaviors will be disabled.
        """
        if enabled:
            print("activating", tag)
        else:
            print("deactivating", tag)
        for obj in self.all_objects():
            for b in obj[tag]:
                b.behavior_enabled = enabled

    def recording_off(self):
        """Turn off recording for all objects in the network."""
        for obj in self.all_objects():
            obj.recording = False

    def recording_on(self):
        """Turn on recording for all objects in the network."""
        for obj in self.all_objects():
            obj.recording = True

    def all_objects(self):
        """Return a list of all objects in the network."""
        l = [self]
        l.extend(self.Structures)
        return l

    def all_behaviors(self):
        """Return a list of all behaviors in the network."""
        result = []
        for obj in self.all_objects():
            for beh in obj.behavior.values():
                result.append(beh)
        return result

    def clear_recorder(self, keys=None):
        """Clear the recorder objects of all network components."""
        for obj in self.all_objects():
            for key in obj.behavior:
                if (keys is None or key in keys) and hasattr(
                    obj.behavior[key], "clear_recorder"
                ):
                    obj.behavior[key].clear_recorder()

    def __repr__(self):
        neuron_count = torch.sum(torch.tensor([ng.size for ng in self.NeuronGroups]))
        synapse_count = torch.sum(
            torch.tensor([sg.src.size * sg.dst.size for sg in self.SynapseGroups if (sg.src is not None and sg.dst is not None)])
        )

        basic_info = (
            "(Neurons: "
            + str(neuron_count)
            + "|"
            + str(len(self.NeuronGroups))
            + " groups, Synapses: "
            + str(synapse_count)
            + "|"
            + str(len(self.SynapseGroups))
            + " groups)"
        )

        result = "Network" + str(self.tags) + basic_info + "{"
        for k in sorted(list(self.behavior.keys())):
            result += str(k) + ":" + str(self.behavior[k])
        result += "}" + "\r\n"

        for ng in self.NeuronGroups:
            result += str(ng) + "\r\n"

        used_tags = []
        for sg in self.SynapseGroups:
            tags = str(sg.tags)
            if tags not in used_tags:
                result += str(sg) + "\r\n"
            used_tags.append(tags)

        for struc in self.Structures:
            if struc not in self.NeuronGroups and struc not in self.SynapseGroups:
                result += str(struc) + "\r\n"

        return result[:-2]

    def find_objects(self, key):
        """Find objects in the network with a specific tag.

        Args:
            key (str): Tag to search for.

        Returns:
            list: List of objects with the tag.
        """
        result = super().find_objects(key)

        for struc in self.Structures:
            result.extend(struc[key])

        for am in self.analysis_modules:
            result.extend(am[key])

        return result

    def initialize(self, info=True, warnings=True, storage_manager=None):
        """Initialize the variables of the network and all its components.

        Args:
            info (bool): If true, print information about the network.
            warnings (bool): If true, print warnings while checking the tag uniqueness.
            storage_manager (StorageManager): Storage manager to use for the network.
        """
        if info:
            desc = str(self)
            print(desc)
            if storage_manager is not None:
                storage_manager.save_param("info", desc)

        self.initialize_behaviors()
        self.check_unique_tags(warnings)
        self.fill_substructures()

    def initialize_behaviors(self):
        for key, parent, behavior in self.sorted_behavior_execution_list:
            if not behavior.initialize_on_init and not behavior.initialize_last:
                behavior.initialize(parent)
                behavior.check_unused_attrs()

        for key, parent, behavior in self.sorted_behavior_execution_list:
            if behavior.initialize_last:
                behavior.initialize(parent)
                behavior.check_unused_attrs()

    def _add_behavior_to_sorted_execution_list(self, key, beh_parent, behavior):
        insert_indx = 0
        for i, kpb in enumerate(self.sorted_behavior_execution_list):
            k, p, b = kpb
            if key >= k:
                insert_indx = i + 1
        self.sorted_behavior_execution_list.insert(
            insert_indx, (key, beh_parent, behavior)
        )

    def _remove_behavior_from_sorted_execution_list(self, key, beh_parent, behavior):
        rm_indx = -1
        for i, kpb in enumerate(self.sorted_behavior_execution_list):
            k, p, b = kpb
            if key == k and beh_parent == p and behavior == b:
                rm_indx = i
                break
        if rm_indx > -1:
            self.sorted_behavior_execution_list.pop(rm_indx)
        else:
            raise Exception("behavior not found")

    def check_unique_tags(self, warnings=True):
        """Check if all tags in the network are unique. In case of doubles, a new tag will be
        automatically assigned to second instance.

        Args:
            warnings (bool): Whether to log the warnings or not.
        """
        unique_tags = []
        for ng in self.NeuronGroups:
            if len(ng.tags) == 0:
                ng.tags.append("NG")
                print('no tag defined for NeuronGroup. "NG" tag added')

            if ng.tags[0] in unique_tags:
                counts = unique_tags.count(ng.tags[0])
                new_tag = ng.tags[0] + chr(97 + counts)
                unique_tags.append(ng.tags[0])
                if warnings:
                    print(
                        'Warning: NeuronGroup Tag "'
                        + ng.tags[0]
                        + '" already in use. The first Tag of an Object should be unique and will be renamed to "'
                        + new_tag
                        + '". Multiple Tags can be separated with a "," (NeuronGroup(..., tag="tag1,tag2,..."))'
                    )
                ng.tags.append(ng.tags[0])
                ng.tags[0] = new_tag

            else:
                unique_tags.append(ng.tags[0])

    def clear_tag_cache(self):
        """Clear the tag cache of all objects in the network for faster search."""
        for obj in self.all_objects():
            obj.clear_cache()

            for k in obj.behavior:
                obj.behavior[k].clear_cache()

    def set_synapses_to_neuron_groups(self):
        """Set the synapses of all synapse groups to the corresponding neuron groups."""
        for ng in self.NeuronGroups:
            ng.afferent_synapses = {"All": []}
            ng.efferent_synapses = {"All": []}

            for sg in self.SynapseGroups:
                for tag in sg.tags:
                    ng.afferent_synapses[tag] = []
                    ng.efferent_synapses[tag] = []

            for sg in self.SynapseGroups:
                if sg.dst.BaseNeuronGroup == ng:
                    for tag in sg.tags + ["All"]:
                        ng.afferent_synapses[tag].append(sg)

                if sg.src.BaseNeuronGroup == ng:
                    for tag in sg.tags + ["All"]:
                        ng.efferent_synapses[tag].append(sg)

    def simulate_iteration(self, measure_behavior_execution_time=False):
        """Simulate one iteration of the network.

        Each iteration includes a `forward` call of objects' behaviors in the order of their keys in the dictionary or list index.

        Args:
            measure_behavior_execution_time (bool): Whether to measure the actual execution time of the behaviors.

        Returns:
            None or dict: If `measure_behavior_execution_time` is set to True, a dictionary with the execution times of the behaviors is returned.
        """
        if measure_behavior_execution_time:
            time_measures = {
                key: 0.0 for key, _, _ in self.sorted_behavior_execution_list
            }

        self.iteration += 1

        for key, parent, behavior in self.sorted_behavior_execution_list:
            if behavior.behavior_enabled and not behavior.empty_iteration_function:
                if measure_behavior_execution_time:
                    start_time = time.time()
                    behavior(parent)
                    time_measures[key] += (time.time() - start_time) * 1000
                else:
                    behavior(parent)

        if measure_behavior_execution_time:
            return time_measures

    def simulate_iterations(
        self,
        iterations,
        batch_size=-1,
        measure_block_time=True,
        disable_recording=False,
        batch_progress_update_func=None,
    ):
        """Simulates the network for a number of iterations.

        Args:
            iterations (int): Number of iterations to simulate.
            batch_size (int): Number of iterations to simulate in one batch. If set to -1, the whole simulation is done in one batch.
            measure_block_time (bool): Whether to measure the time of each batch.
            disable_recording (bool): Whether to disable the recording of the network.
            batch_progress_update_func (function): Function to call after each batch. The function should take the current batch number and network instance as arguments.
        """
        if type(iterations) is str:
            iterations = self["Clock", 0].time_to_iterations(iterations)

        if type(batch_size) is str:
            batch_size = self["Clock", 0].time_to_iterations(batch_size)

        time_diff = None

        if disable_recording:
            self.recording_off()

        if batch_size == -1:
            outside_it = 1
            block_iterations = iterations
        else:
            outside_it = int(iterations / batch_size)
            block_iterations = batch_size

        for t in range(int(outside_it)):
            if measure_block_time:
                start_time = time.time()
            for i in range(int(block_iterations)):
                self.simulate_iteration()
            if measure_block_time:
                time_diff = (time.time() - start_time) * 1000

                print(
                    "\r{}xBatch: {}/{} ({}%) {:.3f}ms".format(
                        block_iterations,
                        t + 1,
                        outside_it,
                        int(100 / outside_it * (t + 1)),
                        time_diff,
                    ),
                    end="",
                )

            if batch_progress_update_func is not None:
                batch_progress_update_func((t + 1.0) / int(outside_it) * 100.0, self)

        for _ in range(iterations % batch_size):
            self.simulate_iteration()

        if disable_recording:
            self.recording_on()

        if measure_block_time:
            print("")

        return time_diff



import torch


def is_number(s):
    """Returns True if string is a number."""
    try:
        float(s)
        return True
    except ValueError:
        return False


def check_is_torch_tensor(x, device="cpu", dtype=torch.float32):
    if not torch.is_tensor(x):
        try:
            return torch.tensor(x, device=device, dtype=dtype)
        except RuntimeError:
            raise RuntimeError("Could not convert NoneTpe to torch tensor: " + str(x))



"""Top-level package for PymoNNtorch."""

__author__ = """Computational Neuroscience Research Laboratory"""
__email__ = "ashenatena@gmail.com"
__version__ = "0.1.4"

from pymonntorch.NetworkCore.Network import *
from pymonntorch.NetworkCore.Behavior import *
from pymonntorch.NetworkCore.NeuronGroup import *
from pymonntorch.NetworkCore.SynapseGroup import *
from pymonntorch.NetworkCore.AnalysisModule import *

from pymonntorch.NetworkBehavior.Basics.BasicHomeostasis import *
from pymonntorch.NetworkBehavior.Basics.Normalization import *

from pymonntorch.NetworkBehavior.EulerEquationModules.Equation import *
from pymonntorch.NetworkBehavior.EulerEquationModules.EulerClock import *
from pymonntorch.NetworkBehavior.EulerEquationModules.Helper import *
from pymonntorch.NetworkBehavior.EulerEquationModules.VariableInitializer import *

from pymonntorch.NetworkBehavior.Structure.Structure import *

from pymonntorch.NetworkBehavior.Recorder.Recorder import *
